/* ************************************************************************** */
// Programmer: Tobin Schwaiger-Hastanan (tobin@13hex.com)
// 
// Copyright (C) 2003 13Hex, Inc.
// All rights reserved.
// 
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions 
// are met:
// 
//      * Redistributions of source code must retain the above copyright 
//        notice, this list of conditions and the following disclaimer.
// 
//      * Redistributions in binary form must reproduce the above 
//        copyright notice, this list of conditions and the following 
//        disclaimer in the documentation and/or other materials 
//        provided with the distribution. 
// 
//      * Neither the name of 13Hex, Inc. nor the names 
//        of its contributors may be used to endorse or promote 
//        products derived from this software without specific prior 
//        written permission. 
// 
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
// LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS 
// FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE 
// COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
// INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
// BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; 
// LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER 
// CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
// LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING 
// IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
// POSSIBILITY OF SUCH DAMAGE.
/* ************************************************************************** */
#include <string>
#include <map>
#include <set>
using namespace std;

#include "snbfrx/entity.h"
#include "snbfrx/entityfactory.h"
#include "snbfrx/entitymanager.h"
using namespace snbfrx;
/** static pointer to an instance of the Entity Manager. */
EntityManager* EntityManager::m_pinstance = NULL;

/**
  * Default constructor, private to protect the singleton implementation.
  */
EntityManager::EntityManager()
{

}

/**
  * Virtual destructor.
  */
EntityManager::~EntityManager()
{

}

/**
  * Singleton instantiation interface.  This method returns
  * a single instance of an entity manager.
  *
  * @return Pointer to an instantiated Entity Manager.
  */
EntityManager* EntityManager::getInstance()
{
    if( m_pinstance == NULL )
        m_pinstance = new EntityManager();

    return m_pinstance;
}

/**
  * Determines if the EntityManager already has a factory registered
  * under a specific name.
  * 
  * @param name string object containing the name to look up.
  * @return bool value specifying if a factory exists by a given name.
  */
bool EntityManager::hasFactory( const string& name ) const
{
    return m_factoryMap.find( name ) != m_factoryMap.end();
}

/**
  * Determines if factory object has already been registered in the EntityManager.          
  * 
  * @param pfactory pointer to an EntityFactory to check existance of.
  * @return bool value specifying if a factory exists.
  */
bool EntityManager::hasFactory( EntityFactory* pfactory ) const
{
    return m_factorySet.find( pfactory ) != m_factorySet.end();
}

/**
  * Registers a factory under a specific name.
  *
  * @param name name of factory being registered.
  * @param pfactory point to factory to be register.
  *
  * @return pointer to the factory added.  If null, factory could not be added. 
  *            user should check to see if a factory exists of the same name or the 
  *         factory object has already be registered.
  */
EntityFactory* EntityManager::registerFactory( const string& name, EntityFactory* pfactory )
{
    if( hasFactory( name ) || hasFactory( pfactory ) )
        return NULL;

    m_factoryMap[ name ] = pfactory;
    m_factorySet.insert( pfactory );

    return pfactory;
}

/**
  * Gets a factory of a specific name.
  *
  * @param name name of factory to get.
  * @return pointer to Entity Factory.  If the factory does not exist, 
  *         null is returned.
  */
EntityFactory* EntityManager::getFactory( const string& name ) const
{
    map<string, EntityFactory*>::const_iterator it = m_factoryMap.find( name );
    return (*it).second;
}

/**
  * Erases a registered entity factory from the manager.
  * This method does not do deletion of the object.
  * 
  * @param name name of the factory to erase.
  * @return point to the factory removed from the manager.
  */
EntityFactory* EntityManager::eraseFactory( const string& name )
{
    EntityFactory* pfactory = NULL;
    
    if( hasFactory( name ) || hasFactory( pfactory ) )
    {
        pfactory = m_factoryMap[ name ];
        m_factoryMap.erase( name );
        m_factorySet.erase( pfactory );
    }

    return pfactory;
}

/**
  * Checks to see if an active entity is already being managed.
  * 
  * @param pentity pointer to entity to check.
  * @return bool value specifying if entity exists or not.
  *         true  - entity exists.
  *         false - entity does not exist.
  */
bool EntityManager::hasEntity( Entity* pentity ) const
{
    return m_entitySet.find( pentity ) != m_entitySet.end();
}

/**
  * Adds an entity to the manager.
  *
  * NOTE: only unique entities may be managed, 
  *       two of the same objects cannot be added.
  * 
  * @param pentity pointer to entity object to be added.
  * @return pointer to entity added. If entity already is managed,
  *         null is returned.
  */
Entity* EntityManager::addEntity( Entity* pentity )
{
    if( hasEntity( pentity ) )
        return NULL;

    m_entitySet.insert( pentity );
    
    return pentity;
}

/**
  * Erases entity from manager.
  * This method does not release the entity, see the update method
  * for more info about releasing entities.
  * 
  * @param pentity pointer to entity to be removed.
  * @return pointer to entity removed from manager.
  */
Entity* EntityManager::eraseEntity( Entity* pentity )
{
    Entity* retval = NULL;
    
    if( hasEntity( pentity ) )
    {
        m_entitySet.erase( pentity );
        retval = pentity;
    }
    
    return retval;
}

/**
  * Updates all active entities.
  */
void EntityManager::update()
{
    set<Entity*> entities( m_entitySet.begin(), m_entitySet.end() );

    // for each active entity...
    for( set<Entity*>::iterator it = entities.begin(); it != entities.end(); it ++ )
    {
        Entity* pentity = *it;

        // update entity. if entity has died, 
        // remove it from the active entity set
        if( !pentity->update() )
        {
            eraseEntity( pentity );
            
            // gets the parent factory and releases the entity 
            // with that factory.
            EntityFactory* pfactory = pentity->getFactory();
            if( pfactory != NULL )
                pfactory->releaseEntity( pentity );
        }
    }
}

/**
  * Renders all active entities.
  */
void EntityManager::render()
{
    // for each active entity...
    for( set<Entity*>::iterator it = m_entitySet.begin(); it != m_entitySet.end(); it ++ )
    {
        Entity* pentity = *it;
        // render entity
        pentity->render();
    }
}